//------------------------
// RooFit Example
//   Convlution of PDFs
//   requires ROOT's FFT support
//
// 2021/06/10
//------------------------
#include "RooRealVar.h"
#include "RooDataSet.h"
#include "RooGaussian.h"
#include "RooLandau.h"
#include "RooFFTConvPdf.h"
#include "RooPlot.h"
#include "TCanvas.h"
#include "TAxis.h"
#include "TH1.h"

using namespace RooFit;

void FitPulse(const char *fname)
{
  //--
  // Define variables
  //--
  double xmin = -1.0;
  double xmax = 10.0;
  RooRealVar t   ("t"   ,"t"   ,      xmin, xmax); // indep variable

  // Fixed vaiables = constants
  RooRealVar n   ("n"   ,"n"   ,       3.);        // order
  RooRealVar tpk ("tpk" ,"tpk" , 120./50.);        // peaking time
  //RooRealVar tpk ("tpk" ,"tpk" , 134./50.);        // peaking time
  //RooRealVar tpk ("tpk" ,"tpk" , 120./50., 110./50., 140./50.);        // peaking time
  RooRealVar mg  ("mg"  ,"mg"  ,       0.);        // bias

  // Fit parameters
  RooRealVar sigt("sigt","sigt",      1.0,   0.,   5.); // diffusion
  RooRealVar t0  ("t0"  ,"t0"  ,     -3.0, -50., 100.); // time offset
  
  //--
  // PDFs
  //--
  RooGenericPdf pulse("pulse", "Generic PDF", 
		      "t-t0>0 ? pow((t-t0)/(tpk/n),n)*exp(-(t-t0)/(tpk/n)) : 0", 
		      RooArgSet(t, t0, tpk, n));
  RooGaussian gauss("gauss", "gauss", t, mg, sigt); // diffusion

  // Construct pulse (x) gauss convoluted PDF
  RooFFTConvPdf pxg("pxg", "pulse (X) gauss", t, pulse, gauss);

  //--
  // Get data
  //--
  cerr   << "Opening " << fname << endl;
  auto hfp = new TFile(fname);

  auto hp = static_cast<TH2D *>(gROOT->FindObject("hPulse3_16"));
  hp->FitSlicesY();
  auto hslp = static_cast<TH1D *>(gROOT->FindObject("hPulse3_16_1"));
  RooDataHist data("data","dataset with tbin",t,hslp) ; 

  //--
  // Fit the convluted PDF to the data
  //--
#if 1
  double xlo = 0.;
  double xhi = 6.;
#else
  double xlo = xmin;
  double xhi = xmax;
#endif
  pxg.fitTo(data, Range(xlo, xhi));

  //--
  // Plot data
  //--
  RooPlot *frame = t.frame(Title("pulse (x) gauss convolution"));
  data.plotOn(frame);                   // plot the data
  pxg.plotOn(frame, LineColor(kRed));   // plot the fit function

  //--
  // Draw frame on canvas
  //--
  new TCanvas("convolution", "convolution", 800, 600);
  gPad->SetLeftMargin(0.15);
  frame->GetYaxis()->SetTitleOffset(1.4);
  frame->Draw();
}
