//------------------------
// RooFit Example
//   Convlution of PDFs
//   requires ROOT's FFT support
//
// 2021/06/24
//------------------------
#include "RooRealVar.h"
#include "RooDataSet.h"
#include "RooGaussian.h"
#include "RooLandau.h"
#include "RooFFTConvPdf.h"
#include "RooPlot.h"
#include "TCanvas.h"
#include "TAxis.h"
#include "TH1.h"

using namespace RooFit;

pair<double, double> FitPulse2(const char *fname, TCanvas *cp = nullptr)
{
  //--
  // Define variables
  //--
#if 1
  double xmin = -1.0;
#else
  double xmin = -2.8;
#endif
  double xmax =  8.0;
  RooRealVar t   ("t"   ,"t"   ,      xmin, xmax); // indep variable

  // Fixed vaiables = constants
  RooRealVar n   ("n"   ,"n"   ,       3.);        // order
  RooRealVar tpk ("tpk" ,"tpk" , 120./50.);        // peaking time
  //RooRealVar tpk ("tpk" ,"tpk" , 134./50.);        // peaking time
  //RooRealVar tpk ("tpk" ,"tpk" , 133./50.);        // peaking time
  RooRealVar mg  ("mg"  ,"mg"  ,       0.);        // bias

  // Fit parameters
  RooRealVar sigt("sigt","sigt",      1.0,   0.,   5.); // diffusion
  RooRealVar t0  ("t0"  ,"t0"  ,     -3.0, -50., 100.); // time offset
  
  //--
  // PDFs
  //--
  RooGenericPdf pulse("pulse", "Generic PDF", 
		      "t-t0>0 ? pow((t-t0)/(tpk/n),n)*exp(-(t-t0)/(tpk/n)) : 0", 
		      RooArgSet(t, t0, tpk, n));
  RooGaussian gauss("gauss", "gauss", t, mg, sigt); // diffusion

  // Construct pulse (x) gauss convoluted PDF
  RooFFTConvPdf pxg("pxg", "pulse (X) gauss", t, pulse, gauss);

  //--
  // Get data
  //--
  cerr   << "Opening " << fname << endl;
  auto hfp = new TFile(fname);

  auto hp = static_cast<TH2D *>(gROOT->FindObject("hPulse3_16"));
  hp->FitSlicesY();
  auto hslp = static_cast<TH1D *>(gROOT->FindObject("hPulse3_16_1"));
  RooDataHist data("data","dataset with tbin",t,hslp) ; 

  //--
  // Fit the convluted PDF to the data
  //--
#if 1
  double xlo = -0.8;
  double xhi = +3.0;
#else
  double xlo = -2.8;
  double xhi = +3.0;
#endif
  t.setRange("signal",xlo,xhi);
#if 0
  pxg.fitTo(data, Range("signal"), SumW2Error(true));
#else
  pxg.chi2FitTo(data, Range("signal"), SumW2Error(true));
#endif

  if (cp) {
    cp->cd();
    //--
    // Plot data
    //--
    RooPlot *frame = t.frame(Title("pulse (x) gauss convolution"));
    data.plotOn(frame);                   // plot the data
    pxg.plotOn(frame, LineColor(kRed));   // plot the fit function

    //--
    // Draw frame on canvas
    //--
    gPad->SetLeftMargin(0.15);
    frame->GetYaxis()->SetTitleOffset(1.4);
    frame->Draw();
  }

  const double kTbinWidth = 50.; // [ns]
  const double kVdrift    = 75.e-3; // [mm/ns]
  double fact = kTbinWidth*kVdrift;

  return pair<double, double>(fact*sigt.getValV(), fact*sigt.getError());
}
